/*--------------------------------------------------------------------------
 *  Copyright 2008 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-shell Project
//
// Maven.java
// Since: Jan 11, 2008
//
// $URL$ 
// $Author$
//--------------------------------------
package org.utgenome.shell;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.xerial.util.StringUtil;
import org.xerial.util.log.Logger;
import org.xerial.util.thread.ThreadManager;

/**
 * Maven utility
 * 
 * @author leo
 * 
 */
public class Maven extends UTGBShellSubCommandBase {

	private static Logger _logger = Logger.getLogger(Maven.class);

	public static boolean isMavenInstalled() {
		String utgbHome = System.getProperty("utgb.home");
		if (utgbHome == null)
			return false;
		else
			return new File(utgbHome, "maven").exists();
	}

	private static String getMavenBinary() {
		String utgbHome = System.getProperty("utgb.home");
		String osName = System.getProperty("os.name");
		if (utgbHome == null || osName == null)
			throw new IllegalStateException("maven is not installed.");
		String mavenStartupScript = (osName.contains("Windows")) ? "mvn.bat" : "mvn";
		return new File(utgbHome, "maven/bin/" + mavenStartupScript).getPath();
	}

	private static abstract class ProcessOutputReader implements Runnable {
		private final BufferedReader reader;

		public ProcessOutputReader(InputStream in) {
			reader = new BufferedReader(new InputStreamReader(in));
		}

		public abstract void output(String line);

		public void run() {
			try {
				String line;
				while ((line = reader.readLine()) != null) {
					output(line);
				}
			}
			catch (IOException e) {
				_logger.error(e);
			}
		}
	}
	
	public static void runMaven(String arg) throws UTGBShellException 
	{
		runMaven(arg, null);
	}

	public static void runMaven(String arg, File workingDir) throws UTGBShellException {
		try {
			ThreadManager threadManager = new ThreadManager(2, 2);
			final Process proc = Runtime.getRuntime().exec(getMavenBinary() + " " + arg, null, workingDir);

			// output the program stdout as
			threadManager.addTask(new ProcessOutputReader(proc.getInputStream()) {
				public void output(String line) {
					_logger.info(line);
				}
			});
			threadManager.addTask(new ProcessOutputReader(proc.getErrorStream()) {
				public void output(String line) {
					_logger.error(line);
				}
			});

			proc.waitFor();
			threadManager.joinAll();

		}
		catch (IOException e) {
			throw new UTGBShellException(e);
		}
		catch (InterruptedException e) {
			throw new UTGBShellException(e);
		}
	}

	public static void explodeWar() throws UTGBShellException {
		runMaven("compile war:exploded");
	}

	@Override
	public void execute(String[] args) throws UTGBShellException {
		runMaven(StringUtil.join(args, " "));
	}

	@Override
	public String name() {
		return "maven";
	}

	public String getDetailedDescription() {
		return loadUsage("help-maven.txt");
	}

	public String getOneLinerDescription() {
		return "execute maven tasks";
	}

	public String getOptionList() {
		return loadUsage("help-maven-options.txt");
	}
}
