/*--------------------------------------------------------------------------
 *  Copyright 2008 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-shell Project
//
// UTGBPortable.java
// Since: Jan 30, 2008
//
// $URL: http://svn.utgenome.org/utgb/trunk/utgb/utgb-shell/src/main/java/org/utgenome/shell/UTGBPortable.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.shell;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Properties;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Link;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.xerial.tomcat.TomcatException;
import org.xerial.tomcat.TomcatServer;
import org.xerial.tomcat.TomcatServerConfiguration;
import org.xerial.util.cui.OptionGroup;
import org.xerial.util.cui.OptionHandler;
import org.xerial.util.cui.OptionParser;
import org.xerial.util.cui.OptionParserException;
import org.xerial.util.thread.ThreadManager;

/**
 * The stand-alone UTGB server launcher
 * 
 * @author leo
 * 
 */
public class UTGBPortable {

	/**
	 * entry point of the UTGBServer
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			UTGBPortable server = new UTGBPortable(args);
			server.start();
		}
		catch (OptionParserException e) {
			System.err.println(e.getMessage());
		}
	}

	private static enum Opt {
		HELP, VERBOSE, PORT, GUI, CUI, CONTEXT_PATH, WEBAPP, WORKING_DIR, CONFIG_DIR
	}

	// configuration parameters
	private static enum LaunchMode {
		CUI, GUI
	}

	private ThreadManager threadManager = new ThreadManager();
	private LaunchMode launchMode = LaunchMode.GUI;
	private int port = 8989;
	private String contextPath = "/utgb";
	private String webappDir = "src/main/webapp";
	private String workingDir = "target/utgb-portable";
	private String configDir = "config";

	/**
	 * Create an instance of the UTGBServer
	 * 
	 * @param args
	 *            command-line arguments
	 * @throws OptionParserException
	 */
	public UTGBPortable(String[] args) throws OptionParserException {
		// setup the command line option parser
		final OptionParser<Opt> opt = new OptionParser<Opt>();
		opt.addOption(Opt.HELP, "h", "help", "display help message", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				// display help messages
				System.err.println(opt.helpMessage());
				throw new OptionParserException("");
			}
		});
		opt.addOption(Opt.VERBOSE, "v", "verbose", "display verbose log messages");
		opt.addOptionWithArgument(Opt.PORT, "p", "port", "PORT_NUMBER", "specify the port number of the local web server (default = 8989)",
				new OptionHandler<Opt>() {
					public void handle(OptionParser<Opt> parser) throws OptionParserException {
						port = parser.getIntValue(Opt.PORT);
					}
				});
		opt.addOptionWithArgument(Opt.CONTEXT_PATH, "c", "path", "PATH", "context path. default=/utgb", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				contextPath = parser.getValue(Opt.CONTEXT_PATH);
			}
		});

		opt.addOptionWithArgument(Opt.WEBAPP, "w", "webapp", "DIR", "webapp directory. default = src/main/webapp", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				webappDir = parser.getValue(Opt.WEBAPP);
			}
		});

		opt.addOptionWithArgument(Opt.WORKING_DIR, "d", "workdir", "DIR", "working directory (tomcat base). default = target/utgb-portable",
				new OptionHandler<Opt>() {
					public void handle(OptionParser<Opt> parser) throws OptionParserException {
						workingDir = parser.getValue(Opt.WORKING_DIR);
					}
				});
		opt.addOptionWithArgument(Opt.CONFIG_DIR, "c", "config", "DIR", "track config directory. default = config", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				configDir = parser.getValue(Opt.CONFIG_DIR);
			}
		});

		// GUI options
		OptionGroup<Opt> guiOptionGroup = new OptionGroup<Opt>("GUI", true);
		guiOptionGroup.addOption(Opt.GUI, "g", "gui", "launch in GUI mode", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				launchMode = LaunchMode.GUI;
			}
		});

		// CUI options
		OptionGroup<Opt> cuiOptionGroup = new OptionGroup<Opt>("CUI", true);
		cuiOptionGroup.addOption(Opt.CUI, "i", "cui", "launch in CUI mode", new OptionHandler<Opt>() {
			public void handle(OptionParser<Opt> parser) throws OptionParserException {
				launchMode = LaunchMode.CUI;
			}
		});

		opt.addOptionGroup(guiOptionGroup);
		opt.addOptionGroup(cuiOptionGroup);

		// parse the command-line arguments
		opt.parse(args);

	}

	/**
	 * Starts the web server
	 */
	public void start() {
		// prepare the config folder
		new File(getConfigDir()).mkdirs();

		switch (launchMode) {
		case GUI:
			runInGUIMode();
			break;
		case CUI:
			runInCUIMode();
			break;
		default:
			throw new IllegalStateException("unknown launch mode");
		}

		try {
			threadManager.joinAll();
		}
		catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	class GUIPanel {
		private final Display display = new Display();
		private final Shell shell = new Shell(display);

		private Label statusLabel;

		private final Composite buttonGroup;
		private final Button startButton;
		private final Button stopButton;
		private final Button restartButton;
		private final Link link;

		private final Group configurationGroup;
		private final Label portFormLabel;
		private final Text portFormText;
		private final Label contextPathLabel;
		private final Text contextPathForm;

		private final Group configSelectionGroup;
		private final Label configFolder;
		private final Button configFolderSelectButton;

		private final Properties properties = new Properties();
		private boolean autostart = false;

		private void loadConfiguration() {
			try {
				BufferedInputStream reader = new BufferedInputStream(new FileInputStream(new File(workingDir, ".config")));
				properties.load(reader);
			}
			catch (FileNotFoundException e) {
				// do nothing
			}
			catch (IOException e) {
				e.printStackTrace();
			}

			// load parameters
			autostart = Boolean.parseBoolean(properties.getProperty("autostart", "false"));
			configDir = properties.getProperty("configDir", "config");
		}

		private void saveConfiguration() {
			BufferedOutputStream writer;
			try {
				File configFileDir = new File(workingDir);
				configFileDir.mkdirs();

				properties.setProperty("autostart", Boolean.toString(autostart));
				properties.setProperty("configDir", configDir);

				writer = new BufferedOutputStream(new FileOutputStream(new File(configFileDir, ".config")));
				properties.store(writer, "UTGB Portable configuration file");
			}
			catch (IOException e) {
				e.printStackTrace();
				setStatus("failed to save configuration");
			}
		}

		public GUIPanel() {
			loadConfiguration();

			// shell window
			shell.setSize(400, 200);
			shell.setText("UTGB Portable");
			shell.setToolTipText("UTGB Portable Server");

			shell.setImage(new Image(display, this.getClass().getResourceAsStream("/org/utgenome/shell/utgb-icon-large.png")));

			shell.setLayout(new RowLayout(SWT.VERTICAL));

			// server start, stop and restart buttons
			buttonGroup = new Composite(shell, SWT.NULL);
			startButton = new Button(buttonGroup, SWT.NONE);
			stopButton = new Button(buttonGroup, SWT.NONE);
			restartButton = new Button(buttonGroup, SWT.NONE);
			link = new Link(buttonGroup, SWT.BOTTOM);

			buttonGroup.setLayout(new GridLayout(4, false));
			startButton.setText("start");
			stopButton.setText("stop");
			restartButton.setText("restart");
			stopButton.setEnabled(false);
			restartButton.setEnabled(false);
			startButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent arg0) {
					startWebServer();
				}
			});
			stopButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent arg0) {
					stopWebServer();
				}
			});

			restartButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent arg0) {
					restartWebServer();
				}
			});

			shell.setDefaultButton(startButton);

			// link button
			link.setText("<a>" + getServerURL() + "</a>");
			link.addSelectionListener(new SelectionAdapter() {

				public void widgetSelected(SelectionEvent arg0) {
					// open the web application from the web browser
					WebBrowser.openURL(getServerURL());
				}
			});

			// auto start button
			Button autostartButton = new Button(shell, SWT.CHECK);
			autostartButton.setText("Automatically start the server on start up");
			autostartButton.setAlignment(SWT.RIGHT);
			autostartButton.setSelection(autostart);
			autostartButton.addSelectionListener(new SelectionAdapter() {

				public void widgetSelected(SelectionEvent e) {
					Button b = (Button) e.widget;
					autostart = b.getSelection();

					saveConfiguration();
				}

			});

			// configuration panel
			configurationGroup = new Group(shell, SWT.EMBEDDED);
			portFormLabel = new Label(configurationGroup, SWT.SHADOW_OUT);
			portFormLabel.setText("port number:");
			portFormText = new Text(configurationGroup, SWT.BORDER | SWT.RIGHT);
			contextPathLabel = new Label(configurationGroup, SWT.NULL);
			contextPathLabel.setText("context path:");
			contextPathForm = new Text(configurationGroup, SWT.BORDER | SWT.LEFT);

			// configuration group
			configurationGroup.setText("server configurations");
			configurationGroup.setLayout(new GridLayout(4, false));

			// text box
			portFormText.setText(Integer.toString(port));
			portFormText.setLayoutData(new GridData(50, SWT.DEFAULT));
			portFormText.setTextLimit(5);
			portFormText.addVerifyListener(new VerifyListener() {
				public void verifyText(VerifyEvent e) {
					int c = e.keyCode;
					if (!(c == 9 || c == 13 || c == 8 || c == 127)) {
						if (!e.text.matches("[0-9]+")) {
							e.doit = false;
						}
					}
				}
			});
			portFormText.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					String inputPort = ((Text) e.widget).getText();
					if (inputPort != null) {
						try {
							port = Integer.parseInt(inputPort);
							updateLink();
						}
						catch (NumberFormatException ne) {
							setStatus("invalid port number");
						}
					}

				}
			});

			contextPathForm.setText(contextPath);
			contextPathForm.setLayoutData(new GridData(80, SWT.DEFAULT));
			contextPathForm.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					setContextPath(((Text) e.widget).getText());
					updateLink();
				}
			});

			// config folder selector
			configSelectionGroup = new Group(shell, SWT.NULL);
			configSelectionGroup.setText("track config folder");
			configSelectionGroup.setLayout(new GridLayout(2, false));
			configFolder = new Label(configSelectionGroup, SWT.LEFT);
			configFolder.setText(getConfigDir());
			configFolderSelectButton = new Button(configSelectionGroup, SWT.PUSH);
			configFolderSelectButton.setText("change");
			configFolderSelectButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					DirectoryDialog dirDialog = new DirectoryDialog(shell);
					dirDialog.setFilterPath(getConfigDir());
					String selectedDir = dirDialog.open();
					if (selectedDir != null) {
						configDir = selectedDir;
						configFolder.setText(getConfigDir());
						saveConfiguration();
						refresh();
					}
				}
			});

			{
				// status bar
				Group statusGroup = new Group(shell, SWT.NONE);
				statusGroup.setLayout(new RowLayout());
				statusGroup.setText("status");
				statusGroup.setSize(shell.getSize().x, SWT.DEFAULT);
				statusLabel = new Label(statusGroup, SWT.NULL);
				statusLabel.setText("server is not started");
				statusLabel.setForeground(display.getSystemColor(SWT.COLOR_DARK_BLUE));
			}

		}

		private void setStatus(String text) {
			statusLabel.setText(text);
			refresh();
		}

		private void updateLink() {
			link.setText("<a>http://localhost:" + port + contextPath + "</a>");
			refresh();
		}

		public void refresh() {
			shell.pack();
			shell.layout();
		}

		public void open() {
			shell.pack();
			shell.open();

			if (autostart) {
				startWebServer();
			}

			while (!shell.isDisposed()) {
				if (!display.readAndDispatch())
					display.sleep();
			}
			display.dispose();
		}

		public void startWebServer() {
			// disable buttons
			startButton.setEnabled(false);
			stopButton.setEnabled(true);
			restartButton.setEnabled(true);

			// disable forms
			portFormText.setEnabled(false);
			contextPathForm.setEnabled(false);

			setStatus("starting the web server...");
			startTomcatServer();
		}

		public void stopWebServer() {
			// disable buttons
			startButton.setEnabled(true);
			stopButton.setEnabled(false);
			restartButton.setEnabled(false);

			// disable forms
			portFormText.setEnabled(true);
			contextPathForm.setEnabled(true);

			setStatus("stopping the web server...");
			stopTomcatServer();
		}

		public void restartWebServer() {
			stopTomcatServer();
			startTomcatServer();
		}

	}

	public void setContextPath(String path) {
		if (!path.startsWith("/"))
			contextPath = "/" + path;
		else
			contextPath = path;
	}

	public String getServerURL() {
		return "http://localhost:" + port + contextPath;
	}

	public String getConfigDir() {
		return new File(configDir).getAbsolutePath();
	}

	protected void runInGUIMode() {
		GUIPanel panel = new GUIPanel();
		panel.open();

	}

	protected void runInCUIMode() {
		startTomcatServer();

		// wait until Ctrl+C terminates the program
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				stopTomcatServer();
			}
		});

		try {
			while (true) {
				Thread.sleep(10000000000000000L);
			}
		}
		catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private static enum ServerStatus {
		STOPPED, STARTED
	}

	private TomcatServer tomcatServer = null;
	private ServerStatus serverStatus = ServerStatus.STOPPED;

	/**
	 * starts the Tomcat server
	 */
	public void startTomcatServer() {
		switch (serverStatus) {
		case STOPPED:
			// create a new instance of the TomcatServer
			TomcatServerConfiguration config = new TomcatServerConfiguration();
			config.setCatalinaBase(workingDir);
			config.setPort(port);
			tomcatServer = new TomcatServer(config);

			try {
				threadManager.addTask(new Runnable() {
					public void run() {
						try {
							// start the server
							tomcatServer.start();
							tomcatServer.addContext(contextPath, new File(webappDir).getAbsolutePath());
						}
						catch (TomcatException e) {
							e.printStackTrace();
						}
						finally {
							serverStatus = ServerStatus.STARTED;
						}
					}
				});
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
			break;
		case STARTED:
			break;
		}

	}

	/**
	 * stops the Tomcat server
	 */
	public void stopTomcatServer() {
		switch (serverStatus) {
		case STOPPED:
			break;
		case STARTED:
			try {
				tomcatServer.stop();
			}
			catch (TomcatException e) {
				e.printStackTrace();
			}
			finally {
				serverStatus = ServerStatus.STOPPED;
			}
			break;
		}

	}

	public String getProjectRoot() {
		return new File(getConfigDir()).getParentFile().getAbsolutePath();
	}

}
